from typing import Generic, TypeVar, Type
from abc import abstractmethod
from citrine._serialization.serializable import Serializable


SelfType = TypeVar('SelfType', bound='Resource')


class PolymorphicSerializable(Generic[SelfType]):
    """A Wrapper class for Polymorphic deserialization of Serializable objects."""

    @classmethod
    @abstractmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Get the underlying type based on given data."""

    @classmethod
    def build(cls, data: dict) -> SelfType:
        """Build the underlying type."""
        subtype = cls.get_type(data)
        return subtype.build(data)



"""Tools for working with workflow resources."""
from typing import Optional
from uuid import UUID

from citrine._rest.asynchronous_object import AsynchronousObject
from citrine._session import Session
from citrine._serialization import properties


__all__ = ['Workflow']


class Workflow(AsynchronousObject):
    """A Citrine Workflow is a collection of Modules that together accomplish some task.

    All workflows must inherit AIResourceMetadata, and hence have a ``status`` field.
    Possible statuses are INPROGRESS, SUCCEEDED, and FAILED.
    Workflows also have a ``status_description`` field with more information.

    """

    _response_key = None
    _session: Optional[Session] = None
    _in_progress_statuses = ["INPROGRESS"]
    _succeeded_statuses = ["SUCCEEDED"]
    _failed_statuses = ["FAILED"]

    project_id: Optional[UUID] = None
    """:Optional[UUID]: Unique ID of the project that contains this workflow."""
    name = properties.String('name')
    description = properties.Optional(properties.String, 'description')
    uid = properties.Optional(properties.UUID, 'id', serializable=False)
    """:Optional[UUID]: Citrine Platform unique identifier"""



from typing import Optional, Set, Type, List

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine.informatics.predictor_evaluation_metrics import PredictorEvaluationMetric

__all__ = ['PredictorEvaluator',
           'CrossValidationEvaluator']


class PredictorEvaluator(PolymorphicSerializable["PredictorEvaluator"]):
    """A Citrine Predictor Evaluator computes metrics on a predictor."""

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        return {
            "CrossValidationEvaluator": CrossValidationEvaluator,
        }[data["type"]]

    def _attrs(self) -> List[str]:
        raise NotImplementedError  # pragma: no cover

    def __eq__(self, other):
        try:
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in self._attrs()
            ])
        except AttributeError:
            return False

    @property
    def responses(self) -> Set[str]:
        """Responses to compute metrics for."""
        raise NotImplementedError  # pragma: no cover

    @property
    def metrics(self) -> Set[PredictorEvaluationMetric]:
        """Metrics to compute for each response."""
        raise NotImplementedError  # pragma: no cover

    @property
    def name(self) -> str:
        """Name of the evaluator.

        A name is required by all evaluators because it is used as the top-level key
        in the results returned by a
        :class:`citrine.informatics.workflows.PredictorEvaluationWorkflow`.
        As such, the names of all evaluators within a single workflow must be unique.
        """
        raise NotImplementedError  # pragma: no cover


class CrossValidationEvaluator(Serializable["CrossValidationEvaluator"], PredictorEvaluator):
    """Evaluate a predictor via cross validation.

    Performs cross-validation on requested predictor responses and computes the requested metrics
    on each response. For a discussion of how many folds and trials to use,
    please see the :ref:`documentation<Cross-validation evaluator>`.

    In addition to a name, set of responses to validate, trials, folds and metrics to compute,
    this evaluator defines a set of descriptor keys to ignore when grouping.  Candidates with
    different values for ignored keys and identical values for all other predictor inputs will be
    placed in the same fold.  For example, if you are baking cakes with different ingredients and
    different oven temperatures and want to group together the data by the ingredients, then you
    can set `ignore_when_grouping={"oven temperature"}`. That way, two recipes that differ only in
    their oven temperature will always end up in the same fold.

    Parameters
    ----------
    name: str
        Name of the evaluator
    description: str
        Description of the evaluator
    responses: Set[str]
        Set of descriptor keys to evaluate
    n_folds: int
        Number of cross-validation folds
    n_trials: int
        Number of cross-validation trials, each contains ``n_folds`` folds
    metrics: Optional[Set[PredictorEvaluationMetric]]
        Optional set of metrics to compute for each response.
        Default is all metrics.
    ignore_when_grouping: Optional[Set[str]]
        Set of descriptor keys to group together.
        Candidates with different values for the given keys and identical values
        for all other descriptors will be in the same group.

    """

    def _attrs(self) -> List[str]:
        return ["typ", "name", "description",
                "responses", "n_folds", "n_trials", "metrics", "ignore_when_grouping"]

    name = properties.String("name")
    description = properties.String("description")
    _responses = properties.Set(properties.String, "responses")
    n_folds = properties.Integer("n_folds")
    n_trials = properties.Integer("n_trials")
    _metrics = properties.Optional(properties.Set(properties.Object(PredictorEvaluationMetric)),
                                   "metrics")
    ignore_when_grouping = properties.Optional(properties.Set(properties.String),
                                               "ignore_when_grouping")
    typ = properties.String("type", default="CrossValidationEvaluator", deserializable=False)

    def __init__(self,
                 name: str, *,
                 description: str = "",
                 responses: Set[str],
                 n_folds: int = 5,
                 n_trials: int = 3,
                 metrics: Optional[Set[PredictorEvaluationMetric]] = None,
                 ignore_when_grouping: Optional[Set[str]] = None):
        self.name: str = name
        self.description: str = description
        self._responses: Set[str] = responses
        self._metrics: Optional[Set[PredictorEvaluationMetric]] = metrics
        self.n_folds: int = n_folds
        self.n_trials: int = n_trials
        self.ignore_when_grouping: Optional[Set[str]] = ignore_when_grouping

    @property
    def responses(self) -> Set[str]:
        """Set of predictor responses cross-validated by the evaluator."""
        return self._responses

    @property
    def metrics(self) -> Set[PredictorEvaluationMetric]:
        """Set of metrics computed during cross-validation."""
        return self._metrics



"""Variable definitions for GEM Tables."""
from abc import abstractmethod
from typing import Type, Optional, List, Union
from uuid import UUID

from deprecation import deprecated

from gemd.entity.bounds.base_bounds import BaseBounds
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.template import ProcessTemplate
from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.template.base_template import BaseTemplate
from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization import properties
from citrine.resources.data_concepts import CITRINE_SCOPE, _make_link_by_uid


class IngredientQuantityDimension(BaseEnumeration):
    """[ALPHA] The dimension of an ingredient quantity.

    * ABSOLUTE corresponds to the absolute quantity
    * MASS corresponds to the mass fraction
    * VOLUME corresponds to the volume fraction
    * NUMBER corresponds to the number fraction
    """

    ABSOLUTE = "absolute"
    MASS = "mass"
    VOLUME = "volume"
    NUMBER = "number"


class DataObjectTypeSelector(BaseEnumeration):
    """[ALPHA] The strategy for selecting types to consider for variable matching.

    Variables can potentially match many objects in a material history, creating
    ambiguity around which value should be assigned. In particular, associated
    runs and specs often share attributes and thus will often match the same
    variable. To enable disambiguation in such circumstances, many variables allow
    specification of a `type_selector`, with the following choices:

    * RUN_ONLY only match run objects
    * SPEC_ONLY only match spec objects
    * PREFER_RUN match either run or spec objects, and if both types match
                 only return the result for runs
    * ANY match either run or spec objects, and if both types match
          return an ambiguous error result
    """

    RUN_ONLY = "run_only"
    SPEC_ONLY = "spec_only"
    PREFER_RUN = "prefer_run"
    ANY = "any"


class Variable(PolymorphicSerializable['Variable']):
    """[ALPHA] A variable that can be assigned values present in material histories.

    Abstract type that returns the proper type given a serialized dict.
    """

    @abstractmethod
    def _attrs(self) -> List[str]:
        pass  # pragma: no cover

    def __eq__(self, other):
        return type(self) == type(other) and all([
            getattr(self, key) == getattr(other, key) for key in self._attrs()
        ])

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        if "type" not in data:
            raise ValueError("Can only get types from dicts with a 'type' key")
        types: List[Type[Serializable]] = [
            TerminalMaterialInfo, AttributeByTemplate, AttributeByTemplateAfterProcessTemplate,
            AttributeByTemplateAndObjectTemplate, IngredientIdentifierByProcessTemplateAndName,
            IngredientLabelByProcessAndName, IngredientLabelsSetByProcessAndName,
            IngredientQuantityByProcessAndName,
            TerminalMaterialIdentifier, AttributeInOutput, IngredientIdentifierInOutput,
            IngredientLabelsSetInOutput, IngredientQuantityInOutput, XOR
        ]
        res = next((x for x in types if x.typ == data["type"]), None)
        if res is None:
            raise ValueError("Unrecognized type: {}".format(data["type"]))

        return res


class TerminalMaterialInfo(Serializable['TerminalMaterialInfo'], Variable):
    """[ALPHA] Metadata from the terminal material of the material history.

    Parameters
    ----------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    field: str
        name of the field to assign the variable to, for example, "sample_type" would
        assign the sample type of the terminal material run

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    field = properties.String('field')
    typ = properties.String('type', default="root_info", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["name", "headers", "field", "typ"]

    def __init__(self,
                 name: str, *,
                 headers: List[str],
                 field: str):
        self.name = name
        self.headers = headers
        self.field = field


@deprecated(deprecated_in="0.133.0", removed_in="2.0.0",
            details="RootInfo is deprecated in favor of TerminalMaterialInfo")
def RootInfo(name: str, *, headers: List[str], field: str) -> TerminalMaterialInfo:
    """[DEPRECATED] Use TerminalMaterialInfo instead."""
    return TerminalMaterialInfo(name=name, headers=headers, field=field)


class AttributeByTemplate(Serializable['AttributeByTemplate'], Variable):
    """[ALPHA] Attribute marked by an attribute template.

    Parameters
    ----------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    template: Union[UUID, str, LinkByUID, AttributeTemplate]
        attribute template that identifies the attribute to assign to the variable
    attribute_constraints: list[list[Union[UUID, str, LinkByUID, AttributeTemplate], Bounds]]
        Optional
        constraints on object attributes in the target object that must be satisfied. Constraints
        are expressed as Bounds.  Attributes are expressed with links. The attribute that the
        variable is being set to may be the target of a constraint as well.
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    template = properties.Object(LinkByUID, 'template')
    attribute_constraints = properties.Optional(
        properties.List(
            properties.SpecifiedMixedList(
                [properties.Object(LinkByUID), properties.Object(BaseBounds)]
            )
        ), 'attribute_constraints')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    typ = properties.String('type', default="attribute_by_template", deserializable=False)

    attribute_type = Union[UUID, str, LinkByUID, AttributeTemplate]
    constraint_type = Union[attribute_type, BaseBounds]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "template", "attribute_constraints", "type_selector", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 template: attribute_type,
                 attribute_constraints: Optional[List[List[constraint_type]]] = None,
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN):
        self.name = name
        self.headers = headers
        self.template = _make_link_by_uid(template)
        self.attribute_constraints = None if attribute_constraints is None \
            else [(_make_link_by_uid(x[0]), x[1]) for x in attribute_constraints]
        self.type_selector = type_selector


class AttributeByTemplateAfterProcessTemplate(
        Serializable['AttributeByTemplateAfterProcessTemplate'], Variable):
    """[ALPHA] Attribute of an object marked by an attribute template and a parent process template.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    attribute_template: Union[UUID, str, LinkByUID, AttributeTemplate]
        attribute template that identifies the attribute to assign to the variable
    process_template: Union[UUID, str, LinkByUID, ProcessTemplate]
        process template that identifies the originating process
    attribute_constraints: list[list[Union[UUID, str, LinkByUID, AttributeTemplate], Bounds]]
        Optional
        constraints on object attributes in the target object that must be satisfied. Constraints
        are expressed as Bounds.  Attributes are expressed with links. The attribute that the
        variable is being set to may be the target of a constraint as well.
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    attribute_template = properties.Object(LinkByUID, 'attribute_template')
    process_template = properties.Object(LinkByUID, 'process_template')
    attribute_constraints = properties.Optional(
        properties.List(
            properties.SpecifiedMixedList(
                [properties.Object(LinkByUID), properties.Object(BaseBounds)]
            )
        ), 'attribute_constraints')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    typ = properties.String('type', default="attribute_after_process", deserializable=False)

    attribute_type = Union[UUID, str, LinkByUID, AttributeTemplate]
    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]
    constraint_type = Union[attribute_type, BaseBounds]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "attribute_template", "process_template",
                "attribute_constraints", "type_selector", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 attribute_template: attribute_type,
                 process_template: process_type,
                 attribute_constraints: Optional[List[List[constraint_type]]] = None,
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN):
        self.name = name
        self.headers = headers
        self.attribute_template = _make_link_by_uid(attribute_template)
        self.process_template = _make_link_by_uid(process_template)
        self.attribute_constraints = None if attribute_constraints is None \
            else [(_make_link_by_uid(x[0]), x[1]) for x in attribute_constraints]
        self.type_selector = type_selector


class AttributeByTemplateAndObjectTemplate(
        Serializable['AttributeByTemplateAndObjectTemplate'], Variable):
    """[ALPHA] Attribute marked by an attribute template and an object template.

    For example, one property may be measured by two different measurement techniques.  In this
    case, that property would have the same attribute template.  Filtering by measurement
    templates, which identify the measurement techniques, disambiguates the technique used to
    measure that otherwise ambiguous property.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    attribute_template: Union[UUID, str, LinkByUID, AttributeTemplate]
        attribute template that identifies the attribute to assign to the variable
    object_template: Union[UUID, str, LinkByUID, BaseTemplate]
        template that identifies the associated object
    attribute_constraints: list[list[Union[UUID, str, LinkByUID, AttributeTemplate], Bounds]]
        Optional
        constraints on object attributes in the target object that must be satisfied. Constraints
        are expressed as Bounds.  Attributes are expressed with links. The attribute that the
        variable is being set to may be the target of a constraint as well.
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    attribute_template = properties.Object(LinkByUID, 'attribute_template')
    object_template = properties.Object(LinkByUID, 'object_template')
    attribute_constraints = properties.Optional(
        properties.List(
            properties.SpecifiedMixedList(
                [properties.Object(LinkByUID), properties.Object(BaseBounds)]
            )
        ), 'attribute_constraints')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    typ = properties.String('type', default="attribute_by_object", deserializable=False)

    attribute_type = Union[UUID, str, LinkByUID, AttributeTemplate]
    object_type = Union[UUID, str, LinkByUID, BaseTemplate]
    constraint_type = Union[attribute_type, BaseBounds]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "attribute_template", "object_template",
                "attribute_constraints", "type_selector", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 attribute_template: attribute_type,
                 object_template: object_type,
                 attribute_constraints: Optional[List[List[constraint_type]]] = None,
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN):
        self.name = name
        self.headers = headers
        self.attribute_template = _make_link_by_uid(attribute_template)
        self.object_template = _make_link_by_uid(object_template)
        self.attribute_constraints = None if attribute_constraints is None \
            else [(_make_link_by_uid(x[0]), x[1]) for x in attribute_constraints]
        self.type_selector = type_selector


class IngredientIdentifierByProcessTemplateAndName(
        Serializable['IngredientIdentifierByProcessAndName'], Variable):
    """[ALPHA] Ingredient identifier associated with a process template and a name.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    process_template: LinkByUID
        process template associated with this ingredient identifier
    ingredient_name: str
        name of ingredient
    scope: str
        scope of the identifier (default: the Citrine scope)
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    process_template = properties.Object(LinkByUID, 'process_template')
    ingredient_name = properties.String('ingredient_name')
    scope = properties.String('scope')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    typ = properties.String('type', default="ing_id_by_process_and_name", deserializable=False)

    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "process_template", "ingredient_name", "scope",
                "type_selector", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 process_template: process_type,
                 ingredient_name: str,
                 scope: str,  # Note that the default is set server side
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN):
        self.name = name
        self.headers = headers
        self.process_template = _make_link_by_uid(process_template)
        self.ingredient_name = ingredient_name
        self.scope = scope
        self.type_selector = type_selector


class IngredientLabelByProcessAndName(Serializable['IngredientLabelByProcessAndName'], Variable):
    """[ALPHA] A boolean variable indicating whether a given label is applied.

    Matches by process template, ingredient name, and the label string to check.

    For example, a column might indicate whether or not the ingredient "ethanol" is labeled as a
    "solvent" in the "second mixing" process.  Many such columns would then support the
    downstream analysis "get the volumetric average density of the solvents".

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    process_template: LinkByUID
        process template associated with this ingredient identifier
    ingredient_name: str
        name of ingredient
    label: str
        label to test
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    process_template = properties.Object(LinkByUID, 'process_template')
    ingredient_name = properties.String('ingredient_name')
    label = properties.String('label')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    typ = properties.String('type', default="ing_label_by_process_and_name", deserializable=False)

    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "process_template", "ingredient_name", "label",
                "type_selector", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 process_template: process_type,
                 ingredient_name: str,
                 label: str,
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN):
        self.name = name
        self.headers = headers
        self.process_template = _make_link_by_uid(process_template)
        self.ingredient_name = ingredient_name
        self.label = label
        self.type_selector = type_selector


class IngredientLabelsSetByProcessAndName(
        Serializable['IngredientLabelsSetByProcessAndName'],
        Variable):
    """[ALPHA] The set of labels on an ingredient when used in a process.

    For example, the ingredient "ethanol" might be labeled "solvent", "alcohol" and "VOC".
    The column would then contain that set of strings.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    process_template: LinkByUID
        process template associated with this ingredient identifier
    ingredient_name: str
        name of ingredient

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    process_template = properties.Object(LinkByUID, 'process_template')
    ingredient_name = properties.String('ingredient_name')
    typ = properties.String('type',
                            default="ing_labels_set_by_process_and_name",
                            deserializable=False)

    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "process_template", "ingredient_name", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 process_template: process_type,
                 ingredient_name: str):
        self.name = name
        self.headers = headers
        self.process_template = _make_link_by_uid(process_template)
        self.ingredient_name = ingredient_name


class IngredientQuantityByProcessAndName(
        Serializable['IngredientQuantityByProcessAndName'], Variable):
    """[ALPHA] The quantity of an ingredient associated with a process template and a name.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    process_template: LinkByUID
        process template associated with this ingredient identifier
    ingredient_name: str
        name of ingredient
    quantity_dimension: IngredientQuantityDimension
        Dimension of the ingredient quantity: absolute quantity, number, mass, or volume fraction.
        Valid options are defined by
        :class:`~citrine.gemtables.variables.IngredientQuantityDimension`
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN
    unit: str
        An optional unit: only ingredient quantities that are convertible to this unit will be
        matched. Note that this parameter is mandatory when quantity_dimension is
        IngredientQuantityDimension.ABSOLUTE.

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    process_template = properties.Object(LinkByUID, 'process_template')
    ingredient_name = properties.String('ingredient_name')
    quantity_dimension = properties.Enumeration(IngredientQuantityDimension, 'quantity_dimension')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    typ = properties.String('type', default="ing_quantity_by_process_and_name",
                            deserializable=False)
    unit = properties.Optional(properties.String, "unit")

    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "process_template", "ingredient_name", "quantity_dimension",
                "type_selector", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 process_template: process_type,
                 ingredient_name: str,
                 quantity_dimension: IngredientQuantityDimension,
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN,
                 unit: Optional[str] = None):
        self.name = name
        self.headers = headers
        self.process_template = _make_link_by_uid(process_template)
        self.ingredient_name = ingredient_name
        self.type_selector = type_selector

        # Cast to make sure the string is valid
        if not isinstance(quantity_dimension, IngredientQuantityDimension):
            quantity_dimension = IngredientQuantityDimension.get_enum(quantity_dimension)
        self.quantity_dimension = quantity_dimension

        if quantity_dimension == IngredientQuantityDimension.ABSOLUTE:
            if unit is None:
                raise ValueError("Absolute Quantity variables require that 'unit' is set")
        else:
            if unit is not None and unit != "":
                raise ValueError("Fractional variables cannot take a 'unit'")
        self.unit = unit


class TerminalMaterialIdentifier(Serializable['TerminalMaterialIdentifier'], Variable):
    """[ALPHA] A unique identifier of the terminal material of the material history, by scope.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    scope: string
        scope of the identifier (default: the Citrine scope)

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    scope = properties.String('scope')
    typ = properties.String('type', default="root_id", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["name", "headers", "scope", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 scope: str = CITRINE_SCOPE):
        self.name = name
        self.headers = headers
        self.scope = scope


@deprecated(deprecated_in="0.133.0", removed_in="2.0.0",
            details="RootIdentifier is deprecated in favor of TerminalMaterialIdentifier")
def RootIdentifier(name: str, *,
                   headers: List[str],
                   scope: str = CITRINE_SCOPE) -> TerminalMaterialIdentifier:
    """[DEPRECATED] Use TerminalMaterialIdentifier instead."""
    return TerminalMaterialIdentifier(name=name, headers=headers, scope=scope)


class AttributeInOutput(Serializable['AttributeInOutput'], Variable):
    """[ALPHA] Attribute marked by an attribute template in the trunk of the history tree.

    The search for an attribute that marks the given attribute template starts at the terminal
    of the material history tree and proceeds until any of the given process templates are reached.
    Those templates block the search from continuing into their ingredients but do not halt the
    search entirely. This variable definition allows attributes that are present both in output
    and the inputs of a process to be distinguished.

    For example, a material "paint" might be produced by mixing and then resting "pigments" and
    a "base".  The color of the pigments and base could be measured and recorded as attributes
    in addition to the color of the resulting paint. To define a variable as the color of the
    resulting paint, AttributeInOutput can be used with the mixing process included in the list
    of process templates. Then, when the platform looks for the color of a paint, it will find it
    but *won't* traverse through the mixing process and also find the colors of the pigments and
    base, which would result in an ambiguous variable match.

    Unlike "AttributeByTemplateAfterProcess", AttributeInOutput will also match on the color
    attribute of the pigments in the rows that correspond to those pigments. This way, all the
    colors can be assigned to the same variable and rendered into the same columns in the GEM
    table.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    attribute_template: LinkByUID
        attribute template that identifies the attribute to assign to the variable
    process_templates: list[LinkByUID]
        process templates that should not be traversed through when searching for a matching
        attribute.  The attribute may be present in these processes but not their ingredients.
    attribute_constraints: list[list[Union[UUID, str, LinkByUID, AttributeTemplate], Bounds]]
        Optional
        constraints on object attributes in the target object that must be satisfied. Constraints
        are expressed as Bounds.  Attributes are expressed with links. The attribute that the
        variable is being set to may be the target of a constraint as well.
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    attribute_template = properties.Object(LinkByUID, 'attribute_template')
    process_templates = properties.List(properties.Object(LinkByUID), 'process_templates')
    attribute_constraints = properties.Optional(
        properties.List(
            properties.SpecifiedMixedList(
                [properties.Object(LinkByUID), properties.Object(BaseBounds)]
            )
        ), 'attribute_constraints')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    typ = properties.String('type', default="attribute_in_trunk", deserializable=False)

    attribute_type = Union[UUID, str, LinkByUID, AttributeTemplate]
    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]
    constraint_type = Union[attribute_type, BaseBounds]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "attribute_template", "process_templates",
                "attribute_constraints", "type_selector", "typ"]

    def __init__(self,
                 name: str,
                 *,
                 headers: List[str],
                 attribute_template: attribute_type,
                 process_templates: List[process_type],
                 attribute_constraints: Optional[List[List[constraint_type]]] = None,
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN):
        self.name = name
        self.headers = headers
        self.attribute_template = _make_link_by_uid(attribute_template)
        self.process_templates = [_make_link_by_uid(x) for x in process_templates]
        self.attribute_constraints = None if attribute_constraints is None \
            else [(_make_link_by_uid(x[0]), x[1]) for x in attribute_constraints]
        self.type_selector = type_selector


class IngredientIdentifierInOutput(Serializable['IngredientIdentifierInOutput'], Variable):
    """[ALPHA] Ingredient identifier in the trunk of a material history tree.

    The search for an ingredient starts at the terminal of the material history tree and
    proceeds until any of the given process templates are reached. Those templates block the search
    from continuing but are inclusive: a match is extracted if an ingredient with the specified
    ingredient name is found at or before a cutoff.

    This variable definition allows an identifier to be extracted when an ingredient is used
    in multiple processes. As an example, consider a paint formed by mixing red and yellow
    pigments, where the red pigment is formed by mixing yellow and magenta. This variable could be
    used to represent the identifier of yellow in both mixing processes (red and the final paint)
    in a single column provided the process templates that mixed red and the final paint
    are included as cutoffs.

    In general, this variable should be preferred over an
    :class:`~citrine.gemtables.variables.IngredientIdentifierByProcessTemplateAndName` when
    mixtures are hierarchical (i.e., blends of blends).
    It allows an ingredient with a single name to be used in
    multiple processes without defining additional variables that manifest as additional columns
    in your GEM table, and must be used in place of the former if the same process template is
    used to represent mixing at multiple levels in the material history hierarchy. Going back
    to the previous example, this variable must be used in place of an
    :class:`~citrine.gemtables.variables.IngredientIdentifierByProcessTemplateAndName` if the same
    process template was used to represent the process that mixed red and the final paint.
    Using :class:`~citrine.gemtables.variables.IngredientIdentifierByProcessTemplateAndName`
    would result in an ambiguous match because yellow would be found twice in the
    material history, once when mixing red and again when mixing the final paint.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    ingredient_name: str
        Name of the ingredient to search for
    process_templates: list[Union[UUID, str, LinkByUID, ProcessTemplate]]
        Process templates halt the search for a matching ingredient name.
        These process templates are inclusive.
        The ingredient may be present in these processes but not before.
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    ingredient_name = properties.String('ingredient_name')
    process_templates = properties.List(properties.Object(LinkByUID), 'process_templates')
    scope = properties.String('scope')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    typ = properties.String('type', default="ing_id_in_output", deserializable=False)

    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "ingredient_name", "process_templates", "scope",
                "type_selector", "typ"]

    def __init__(self,
                 name: str, *,
                 headers: List[str],
                 ingredient_name: str,
                 process_templates: List[process_type],
                 scope: str = CITRINE_SCOPE,
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN):
        self.name = name
        self.headers = headers
        self.ingredient_name = ingredient_name
        self.process_templates = [_make_link_by_uid(x) for x in process_templates]
        self.scope = scope
        self.type_selector = type_selector


class IngredientLabelsSetInOutput(Serializable['IngredientLabelsSetInOutput'], Variable):
    """[ALPHA] The set of labels on an ingredient in the trunk of a material history tree.

    The search for an ingredient starts at the terminal of the material history tree and proceeds
    until any of the given process templates are reached. Those templates block the search from
    continuing but are inclusive: a match is extracted if an ingredient with the specified
    ingredient name is found at or before a cutoff.

    This variable definition allows a set of labels to be extracted when an ingredient is used
    in multiple processes. As an example, consider a paint formed by mixing red and yellow
    pigments, where the red pigment is formed by mixing yellow and magenta. This variable could be
    used to represent the labels applied to yellow in both mixing processes (red and the final
    paint) in a single column provided the process templates that mixed red and the final paint
    are included as cutoffs.

    In general, this variable should be preferred over an
    :class:`~citrine.gemtables.variables.IngredientLabelsSetByProcessAndName` when
    mixtures are hierarchical (i.e., blends of blends).
    It allows an ingredient with a single name to be used in
    multiple processes without defining additional variables that manifest as additional columns
    in your GEM table, and must be used in place of the former if the same process template is
    used to represent mixing at multiple levels in the material history hierarchy. Going back
    to the previous example, this variable must be used in place of an
    :class:`~citrine.gemtables.variables.IngredientLabelsSetByProcessAndName` if the same
    process template was used to represent the process that mixed red and the final paint.
    Using :class:`~citrine.gemtables.variables.IngredientLabelsSetByProcessAndName`
    would result in an ambiguous match because yellow would be found twice in the
    material history, once when mixing red and again when mixing the final paint.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    process_templates: list[Union[UUID, str, LinkByUID, ProcessTemplate]]
        process templates that should not be traversed through when searching for a matching
        attribute.  The attribute may be present in these processes but not their ingredients.
    ingredient_name: str
        name of ingredient

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    process_templates = properties.List(properties.Object(LinkByUID), 'process_templates')
    ingredient_name = properties.String('ingredient_name')
    typ = properties.String('type', default="ing_label_set_in_output", deserializable=False)

    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]

    def _attrs(self) -> List[str]:
        return ["name", "headers", "process_templates", "ingredient_name", "typ"]

    def __init__(self,
                 name: str, *,
                 headers: List[str],
                 process_templates: List[process_type],
                 ingredient_name: str):
        self.name = name
        self.headers = headers
        self.process_templates = [_make_link_by_uid(x) for x in process_templates]
        self.ingredient_name = ingredient_name


class IngredientQuantityInOutput(Serializable['IngredientQuantityInOutput'], Variable):
    """[ALPHA] Ingredient quantity in the trunk of a material history tree.

    The search for an ingredient starts at the terminal of the material history tree and proceeds
    until any of the given process templates are reached. Those templates block the search from
    continuing but are inclusive: a match is extracted if an ingredient with the specified
    ingredient name is found at or before a cutoff.

    This variable definition allows a quantity to be extracted when an ingredient is used in
    multiple processes. As an example, consider a paint formed by mixing red and yellow pigments,
    where the red pigment is formed by mixing yellow and magenta. This variable could be used to
    represent the quantity of yellow in both mixing processes (red and the final paint) in a
    single column provided the process templates that mixed red and the final paint
    are included as cutoffs.

    In general, this variable should be preferred over an
    :class:`~citrine.gemtables.variables.IngredientQuantityByProcessTemplateAndName`
    when mixtures are hierarchical (i.e., blends of blends). It allows an ingredient with a
    single name to be used in multiple processes without defining additional variables
    that manifest as additional columns in your table, and must be used in place of the
    former if the same process template is used to represent mixing at multiple levels
    in the material history hierarchy.
    Going back to the previous example, this variable must be used in place of an
    :class:`~citrine.gemtables.variables.IngredientQuantityByProcessTemplateAndName` if the same
    process template was used to represent the process that mixed red and the final paint.
    Using :class:`~citrine.gemtables.variables.IngredientQuantityByProcessTemplateAndName`
    would result in an ambiguous match because yellow would be found twice in the material history,
    once when mixing red and again when mixing the final paint.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    ingredient_name: str
        Name of the ingredient to search for
    quantity_dimension: IngredientQuantityDimension
        Dimension of the ingredient quantity: absolute quantity, number, mass, or volume fraction.
        Valid options are defined by
        :class:`~citrine.gemtables.variables.IngredientQuantityDimension`
    process_templates: list[Union[UUID, str, LinkByUID, ProcessTemplate]]
        Process templates halt the search for a matching ingredient name.
        These process templates are inclusive.
        The ingredient may be present in these processes but not before.
    type_selector: DataObjectTypeSelector
        strategy for selecting data object types to consider when matching, defaults to PREFER_RUN
    unit: str
        an optional unit: only ingredient quantities that are convertible to this unit will be
        matched. note that this parameter is mandatory when quantity_dimension is
        IngredientQuantityDimension.ABSOLUTE.

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    ingredient_name = properties.String('ingredient_name')
    quantity_dimension = properties.Enumeration(IngredientQuantityDimension, 'quantity_dimension')
    process_templates = properties.List(properties.Object(LinkByUID), 'process_templates')
    type_selector = properties.Enumeration(DataObjectTypeSelector, "type_selector")
    unit = properties.Optional(properties.String, "unit")
    typ = properties.String('type', default="ing_quantity_in_output", deserializable=False)

    process_type = Union[UUID, str, LinkByUID, ProcessTemplate]

    def _attrs(self) -> List[str]:
        return ["name",
                "headers",
                "ingredient_name",
                "process_templates",
                "type_selector",
                "unit",
                "typ"]

    def __init__(self,
                 name: str, *,
                 headers: List[str],
                 ingredient_name: str,
                 quantity_dimension: IngredientQuantityDimension,
                 process_templates: List[process_type],
                 type_selector: DataObjectTypeSelector = DataObjectTypeSelector.PREFER_RUN,
                 unit: Optional[str] = None):
        self.name = name
        self.headers = headers
        self.ingredient_name = ingredient_name
        self.process_templates = [_make_link_by_uid(x) for x in process_templates]
        self.type_selector = type_selector

        # Cast to make sure the string is valid
        if not isinstance(quantity_dimension, IngredientQuantityDimension):
            quantity_dimension = IngredientQuantityDimension.get_enum(quantity_dimension)
        self.quantity_dimension = quantity_dimension

        if quantity_dimension == IngredientQuantityDimension.ABSOLUTE:
            if unit is None:
                raise ValueError("Absolute Quantity variables require that 'unit' is set")
        else:
            if unit is not None and unit != "":
                raise ValueError("Fractional variables cannot take a 'unit'")
        self.unit = unit


class XOR(Serializable['XOR'], Variable):
    """[ALPHA] Logical exclusive OR for GEM table variables.

    This variable combines the results of 2 or more variables into a single variable according to
    exclusive OR logic. XOR is defined when exactly one of its inputs is defined. Otherwise it is
    undefined.

    XOR can only operate on inputs with the same output type. For example, you may XOR
    :class:`~citrine.gemtables.variables.TerminalMaterialIdentifier` with
    :class:`~citrine.gemtables.variables.IngredientIdentifierByProcessTemplateAndName`
    because they both produce simple strings, but not with
    :class:`~citrine.gemtables.variables.IngredientQuantityInOutput`
    which produces a real numeric quantity.

    The input variables to XOR need not exist elsewhere in the table config, and the name and
    headers assigned to them have no bearing on how the table is constructed. That they are
    required at all is simply a limitation of the current API.

    Parameters
    ---------
    name: str
        a short human-readable name to use when referencing the variable
    headers: list[str]
        sequence of column headers
    variables: list[Variable]
        set of 2 or more Variables to XOR

    """

    name = properties.String('name')
    headers = properties.List(properties.String, 'headers')
    variables = properties.List(properties.Object(Variable), 'variables')
    typ = properties.String('type', default="xor", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["name", "headers", "variables", "typ"]

    def __init__(self, name, *, headers, variables):
        self.name = name
        self.headers = headers
        self.variables = variables



"""Collection class for generic GEMD objects and templates."""
from typing import Type, Union, Optional, List, Tuple, Iterable
from uuid import UUID, uuid4
import re
from tqdm.auto import tqdm

from gemd.entity.base_entity import BaseEntity
from gemd.entity.link_by_uid import LinkByUID
from gemd.util import recursive_flatmap, recursive_foreach, set_uuids, \
    make_index, substitute_objects

from citrine.resources.api_error import ApiError
from citrine.resources.data_concepts import DataConcepts, DataConceptsCollection, \
    CITRINE_SCOPE, CITRINE_TAG_PREFIX
from citrine.resources.delete import _async_gemd_batch_delete
from citrine._session import Session
from citrine._utils.functions import scrub_none, replace_objects_with_links


class GEMDResourceCollection(DataConceptsCollection[DataConcepts]):
    """A collection of any kind of GEMD objects/templates."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/storables'
    _dataset_agnostic_path_template = 'projects/{project_id}/storables'

    def __init__(self, project_id: UUID, dataset_id: UUID, session: Session):
        DataConceptsCollection.__init__(self,
                                        project_id=project_id,
                                        dataset_id=dataset_id,
                                        session=session)
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.session = session

    @classmethod
    def get_type(cls) -> Type[DataConcepts]:
        """Return the resource type in the collection."""
        return DataConcepts

    def _collection_for(self, model):
        collection = DataConcepts.get_collection_type(model)
        return collection(self.project_id, self.dataset_id, self.session)

    def build(self, data: dict) -> DataConcepts:
        """
        Build an arbitary GEMD object from a serialized dictionary.

        This is an internal method, and should not be called directly by users.

        Parameters
        ----------
        data: dict
            A serialized data model object.

        Returns
        -------
        DataConcepts
            A data model object built from the dictionary.

        """
        return super().build(data)

    def register_all(self,
                     models: Iterable[DataConcepts],
                     *,
                     dry_run=False,
                     status_bar=False,
                     include_nested=False) -> List[DataConcepts]:
        """
        Register multiple GEMD objects to each of their appropriate collections.

        Does so in an order that is guaranteed to store all linked items before the item that
        references them.

        If the GEMD objects have no UIDs, Citrine IDs will be assigned to them prior to passing
        them on to the server.  This is required as otherwise there is no way to determine how
        objects are related to each other.  When the registered objects are returned from the
        server, the input GEMD objects will be updated with whichever uids & _citr_auto:: tags are
        on the returned objects.  This means GEMD objects that already exist on the server will
        be updated with all their on-platform uids and tags.

        This method has the same behavior as `register`, except that no models will be
        written if any one of them is invalid.  Using this method should yield significant
        improvements to write speed over separate calls to `register`.

        Parameters
        ----------
        models: Iterable[DataConcepts]
            The data model objects to register. Can be different types.

        dry_run: bool
            Whether to actually register the item or run a dry run of the register operation.
            Dry run is intended to be used for validation. Default: false

        status_bar: bool
            Whether to display a status bar using the tqdm module to track progress in
            registration. Requires installing the optional tqdm module. Default: false

        include_nested: bool
            Whether to just register the objects passed in the list, or include nested objects
            (e.g., obj.process, obj.spec.template, ...).  Default: false

        Returns
        -------
        List[DataConcepts]
            The registered versions

        """
        from citrine._utils.batcher import Batcher

        if self.dataset_id is None:
            raise RuntimeError("Must specify a dataset in order to register a data model object.")
        path = self._get_path()
        params = {'dry_run': dry_run}

        if include_nested:
            models = recursive_flatmap(models, lambda o: [o], unidirectional=False)

        temp_scope = str(uuid4())
        scope = temp_scope if dry_run else CITRINE_SCOPE
        set_uuids(models, scope=scope)

        resources = list()
        batch_size = 50
        result_index = dict()
        if dry_run:
            batcher = Batcher.by_dependency()
        else:
            batcher = Batcher.by_type()

        if status_bar:
            desc = "Verifying GEMDs" if dry_run else "Registering GEMDs"
            iterator = tqdm(batcher.batch(models, batch_size), leave=False, desc=desc)
        else:
            iterator = batcher.batch(models, batch_size)

        for batch in iterator:
            objects = [replace_objects_with_links(scrub_none(model.dump())) for model in batch]
            response_data = self.session.put_resource(
                path + '/batch',
                json={'objects': objects},
                params=params
            )
            registered = [self.build(obj) for obj in response_data['objects']]
            result_index.update(make_index(registered))
            substitute_objects(registered, result_index, inplace=True)

            if not dry_run:
                # Platform may add a CITRINE_SCOPE uid and citr_auto tags; update locals
                for obj in batch:
                    result = result_index[obj.to_link()]
                    obj.uids.update({k: v for k, v in result.uids.items()})
                    if result.tags is not None:
                        obj.tags = list(result.tags)
            else:
                # Remove of the tags/uids the platform spuriously added
                # this might leave objects with just the temp ids, which we want to strip later
                for obj in batch:
                    result = result_index[obj.to_link()]
                    if CITRINE_SCOPE not in obj.uids:
                        citr_id = result.uids.pop(CITRINE_SCOPE, None)
                        result_index.pop(LinkByUID(scope=CITRINE_SCOPE, id=citr_id), None)
                    if result.tags is not None:
                        todo = [tag for tag in result.tags
                                if re.match(f"^{CITRINE_TAG_PREFIX}::", tag)]
                        for tag in todo:  # Covering this block would require dark art
                            if tag not in obj.tags:
                                result.tags.remove(tag)

            resources.extend(registered)

        if dry_run:  # No-op if not dry-run
            recursive_foreach(list(models) + list(resources),
                              lambda x: x.uids.pop(temp_scope, None))  # Strip temp uids
        return resources

    def async_update(self, model: DataConcepts, *,
                     dry_run: bool = False,
                     wait_for_response: bool = True,
                     timeout: float = 2 * 60,
                     polling_delay: float = 1.0,
                     return_model: bool = False) -> Optional[Union[UUID, DataConcepts]]:
        """
        [ALPHA] Update a particular element of the collection with data validation.

        Update a particular element of the collection, doing a deeper check to ensure that
        the dependent data objects are still with the (potentially) changed constraints
        of this change. This will allow you to make bounds and allowed named/labels changes
        to templates.

        Parameters
        ----------
        model: DataConcepts
            The DataConcepts object.
        dry_run: bool
            Whether to actually update the item or run a dry run of the update operation.
            Dry run is intended to be used for validation. Default: false
        wait_for_response: bool
            Whether to poll for the eventual response. This changes the return type (see
            below).
        timeout: float
            How long to poll for the result before giving up. This is expressed in
            (fractional) seconds.
        polling_delay: float
            How long to delay between each polling retry attempt.
        return_model: bool
            Whether or not to return an updated version of the resource
            If wait_for_response is False, then this argument has no effect

        Returns
        -------
        Optional[UUID]
            If wait_for_response if True, then this call will poll the backend, waiting
            for the eventual job result. In the case of successful validation/update,
            a return value of None is provided unless return_model is True, in which case
            the updated resource is fetched and returned. In the case of a failure
            validating or processing the update, an exception (JobFailureError) is raised
            and an error message is logged with the underlying reason of the failure.

            If wait_for_response if False, A job ID (of type UUID) is returned that one
            can use to poll for the job completion and result with the
            :func:`~citrine.resources.DataConceptsCollection.poll_async_update_job`
            method.

        """
        return self._collection_for(model).async_update(
            model,
            dry_run=dry_run,
            wait_for_response=wait_for_response,
            timeout=timeout,
            polling_delay=polling_delay,
            return_model=return_model
        )

    def batch_delete(
            self,
            id_list: List[Union[LinkByUID, UUID, str, BaseEntity]],
            *,
            timeout: float = 2 * 60,
            polling_delay: float = 1.0
    ) -> List[Tuple[LinkByUID, ApiError]]:
        """
        Remove a set of GEMD objects.

        You may provide GEMD objects that reference each other, and the objects
        will be removed in the appropriate order.

        A failure will be returned if the object cannot be deleted due to an external
        reference.

        All data objects must be associated with this dataset resource. You must also
        have write access on this dataset.

        If you wish to delete more than 50 objects, queuing of deletes requires that
        the types of objects be known, and thus you _must_ provide ids in the form
        of BaseEntities.

        Also note that Attribute Templates cannot be deleted at present.

        Parameters
        ----------
        id_list: List[Union[LinkByUID, UUID, str, BaseEntity]]
            A list of the IDs of data objects to be removed. They can be passed
            as a LinkByUID tuple, a UUID, a string, or the object itself. A UUID
            or string is assumed to be a Citrine ID, whereas a LinkByUID or
            BaseEntity can also be used to provide an external ID.

        Returns
        -------
        List[Tuple[LinkByUID, ApiError]]
            A list of (LinkByUID, api_error) for each failure to delete an object.
            Note that this method doesn't raise an exception if an object fails to be
            deleted.

        """
        return _async_gemd_batch_delete(id_list, self.project_id, self.session, self.dataset_id,
                                        timeout=timeout, polling_delay=polling_delay)



from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.template.base_template import BaseTemplate


def split_templates_from_objects(gems):
    """
    Sort the provided gems into two lists.

    The first list contains only templates (both ObjectTemplate and AttributeTemplate).
    The second list contains only data objects (specs and runs).

    Useful when the procedure for seeding templates needs to be distinct from the procedure for
    seeding data objects.

    Note that each list is unsorted and it is expected that subsequent registration code will
    handle sorting the gems provided.
    """
    templates = []
    data_objects = []
    for gem in gems:
        if isinstance(gem, AttributeTemplate) or isinstance(gem, BaseTemplate):
            templates.append(gem)
        else:
            data_objects.append(gem)
    return templates, data_objects



"""Tools for working with Descriptors."""
from abc import abstractmethod
from typing import Type, List, Mapping, Optional, Union
from uuid import UUID

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine.informatics.descriptors import Descriptor, FormulationDescriptor
from citrine.resources.file_link import FileLink

__all__ = ['DataSource',
           'CSVDataSource',
           'GemTableDataSource']


class DataSource(PolymorphicSerializable['DataSource']):
    """A source of data for the AI engine.

    Data source provides a polymorphic interface for specifying different kinds of data as the
    training data for predictors and the input data for some design spaces.

    """

    @abstractmethod
    def _attrs(self) -> List[str]:
        pass  # pragma: no cover

    def __eq__(self, other):
        try:
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in self._attrs()
            ])
        except AttributeError:
            return False

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        if "type" not in data:
            raise ValueError("Can only get types from dicts with a 'type' key")
        types: List[Type[Serializable]] = [
            CSVDataSource, GemTableDataSource
        ]
        res = next((x for x in types if x.typ == data["type"]), None)
        if res is None:
            raise ValueError("Unrecognized type: {}".format(data["type"]))
        return res


class CSVDataSource(Serializable['CSVDataSource'], DataSource):
    """A data source based on a CSV file stored on the data platform.

    Parameters
    ----------
    file_link: FileLink
        link to the CSV file to read the data from
    column_definitions: Mapping[str, Descriptor]
        Map the column headers to the descriptors that will be used to interpret the cell contents
    identifiers: Optional[List[str]]
        List of one or more column headers whose values uniquely identify a row. These may overlap
        with ``column_definitions`` if a column should be used as data and as an identifier,
        but this is not necessary. Identifiers must be unique within a dataset. No two rows can
        contain the same value.

    """

    typ = properties.String('type', default='csv_data_source', deserializable=False)
    file_link = properties.Object(FileLink, "file_link")
    column_definitions = properties.Mapping(
        properties.String, properties.Object(Descriptor), "column_definitions")
    identifiers = properties.Optional(properties.List(properties.String), "identifiers")

    def _attrs(self) -> List[str]:
        return ["file_link", "column_definitions", "identifiers", "typ"]

    def __init__(self,
                 *,
                 file_link: FileLink,
                 column_definitions: Mapping[str, Descriptor],
                 identifiers: Optional[List[str]] = None):
        self.file_link = file_link
        self.column_definitions = column_definitions
        self.identifiers = identifiers


class GemTableDataSource(Serializable['GemTableDataSource'], DataSource):
    """A data source based on a GEM Table hosted on the data platform.

    Parameters
    ----------
    table_id: UUID
        Unique identifier for the GEM Table
    table_version: Union[str,int]
        Version number for the GEM Table. The first GEM table built from a configuration
        has version = 1. Strings are cast to ints.
    formulation_descriptor: Optional[FormulationDescriptor]
        Optional descriptor used to store formulations emitted by the data source.
        If the data source emits a formulation but this argument is not provided, then a
        default formulation descriptor will be generated. The formulations descriptor, and
        other descriptors, can be retrieved using
        :func:`~citrine.resources.descriptors.DescriptorMethods.descriptors_from_data_source`.

    """

    typ = properties.String('type', default='hosted_table_data_source', deserializable=False)
    table_id = properties.UUID("table_id")
    table_version = properties.Integer("table_version")
    formulation_descriptor = properties.Optional(
        properties.Object(FormulationDescriptor),
        "formulation_descriptor"
    )

    def _attrs(self) -> List[str]:
        return ["table_id", "table_version", "typ"]

    def __init__(self,
                 *,
                 table_id: UUID,
                 table_version: Union[int, str],
                 formulation_descriptor: Optional[FormulationDescriptor] = None):
        self.table_id: UUID = table_id
        self.table_version: Union[int, str] = table_version
        self.formulation_descriptor: Optional[FormulationDescriptor] = formulation_descriptor



from typing import List, Optional
from warnings import warn

from citrine._rest.engine_resource import EngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.data_sources import DataSource
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.predictors import Predictor

__all__ = ['SimpleMLPredictor']


class SimpleMLPredictor(EngineResource['SimplePredictor'], Predictor):
    """[DEPRECATED] A predictor interface that builds a simple graphical model.

    The model connects the set of inputs through latent variables to the outputs.
    Supported complex inputs (such as chemical formulas) are auto-featurized and machine learning
    models are built for each latent variable and output.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        the description of the predictor
    inputs: list[Descriptor]
        Descriptors that represent inputs to relations
    outputs: list[Descriptor]
        Descriptors that represent outputs of relations
    latent_variables: list[Descriptor]
        Descriptors that are predicted from inputs and used when predicting the outputs
    training_data: Optional[List[DataSource]]
        Sources of training data. Each can be either a CSV or an GEM Table. Candidates from
        multiple data sources will be combined into a flattened list and de-duplicated by uid and
        identifiers. de-duplication is performed if a uid or identifier is shared between two or
        more rows. The content of a de-duplicated row will contain the union of data across all
        rows that share the same uid or at least 1 identifier. Training data is unnecessary if the
        predictor is part of a graph that includes all training data required by this predictor.

    """

    inputs = _properties.List(_properties.Object(Descriptor), 'data.instance.inputs')
    outputs = _properties.List(_properties.Object(Descriptor), 'data.instance.outputs')
    latent_variables = _properties.List(_properties.Object(Descriptor),
                                        'data.instance.latent_variables')
    training_data = _properties.List(_properties.Object(DataSource),
                                     'data.instance.training_data', default=[])

    typ = _properties.String('data.instance.type', default='Simple', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 inputs: List[Descriptor],
                 outputs: List[Descriptor],
                 latent_variables: List[Descriptor],
                 training_data: Optional[List[DataSource]] = None):
        warn("SimpleMLPredictor is deprecated as of 1.29.0 and will be removed in 2.0.0."
             " Please use the builders.predictors.build_simple_ml.", category=DeprecationWarning)
        self.name: str = name
        self.description: str = description
        self.inputs: List[Descriptor] = inputs
        self.outputs: List[Descriptor] = outputs
        self.latent_variables: List[Descriptor] = latent_variables
        self.training_data: List[DataSource] = training_data or []

    def __str__(self):
        return '<SimplePredictor {!r}>'.format(self.name)



"""Resources that represent process templates."""
from typing import List, Dict, Optional, Union, Sequence, Type

from citrine._rest.resource import Resource
from citrine._serialization.properties import List as PropertyList
from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import Union as PropertyUnion
from citrine._serialization.properties import String, Mapping, Object, SpecifiedMixedList, \
    LinkOrElse
from citrine.resources.condition_template import ConditionTemplate
from citrine.resources.data_concepts import DataConcepts
from citrine.resources.object_templates import ObjectTemplate, ObjectTemplateCollection
from citrine.resources.parameter_template import ParameterTemplate
from gemd.entity.bounds.base_bounds import BaseBounds
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.template.process_template import ProcessTemplate as GEMDProcessTemplate


class ProcessTemplate(ObjectTemplate, Resource['ProcessTemplate'], GEMDProcessTemplate):
    """
    A process template.

    Process templates are collections of condition and parameter templates that constrain the
    values of a measurement's condition and parameter attributes, and provide a common structure
    for describing similar measurements.

    Parameters
    ----------
    name: str
        The name of the process template.
    description: str, optional
        Long-form description of the process template.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/gemd-docs/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/gemd-docs/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    conditions: List[ConditionTemplate] or List[ConditionTemplate, \
    :py:class:`BaseBounds <gemd.entity.bounds.base_bounds.BaseBounds>`], optional
        Templates for associated conditions. Each template can be provided by itself, or as a list
        with the second entry being a separate, *more restrictive* Bounds object that defines
        the limits of the value for this condition.
    parameters: List[ParameterTemplate] or List[ParameterTemplate, \
    :py:class:`BaseBounds <gemd.entity.bounds.base_bounds.BaseBounds>`], optional
        Templates for associated parameters. Each template can be provided by itself, or as a list
        with the second entry being a separate, *more restrictive* Bounds object that defines
        the limits of the value for this parameter.

    """

    _response_key = GEMDProcessTemplate.typ  # 'process_template'

    name = String('name', override=True)
    description = PropertyOptional(String(), 'description', override=True)
    uids = Mapping(String('scope'), String('id'), 'uids', override=True)
    tags = PropertyOptional(PropertyList(String()), 'tags', override=True)
    conditions = PropertyOptional(PropertyList(PropertyUnion([LinkOrElse, SpecifiedMixedList(
        [LinkOrElse, PropertyOptional(Object(BaseBounds))])])), 'conditions', override=True)
    parameters = PropertyOptional(PropertyList(PropertyUnion([LinkOrElse, SpecifiedMixedList(
        [LinkOrElse, PropertyOptional(Object(BaseBounds))])])), 'parameters', override=True)
    allowed_labels = PropertyOptional(PropertyList(String()), 'allowed_labels', override=True)
    allowed_names = PropertyOptional(PropertyList(String()), 'allowed_names', override=True)
    typ = String('type')

    def __init__(self,
                 name: str,
                 *,
                 uids: Optional[Dict[str, str]] = None,
                 conditions: Optional[Sequence[Union[ConditionTemplate,
                                                     LinkByUID,
                                                     Sequence[Union[ConditionTemplate, LinkByUID,
                                                                    Optional[BaseBounds]]]
                                                     ]]] = None,
                 parameters: Optional[Sequence[Union[ParameterTemplate,
                                                     LinkByUID,
                                                     Sequence[Union[ParameterTemplate, LinkByUID,
                                                                    Optional[BaseBounds]]]
                                                     ]]] = None,
                 allowed_labels: Optional[List[str]] = None,
                 allowed_names: Optional[List[str]] = None,
                 description: Optional[str] = None,
                 tags: Optional[List[str]] = None):
        if uids is None:
            uids = dict()
        DataConcepts.__init__(self, GEMDProcessTemplate.typ)
        GEMDProcessTemplate.__init__(self, name=name, uids=uids,
                                     conditions=conditions, parameters=parameters, tags=tags,
                                     description=description, allowed_labels=allowed_labels,
                                     allowed_names=allowed_names)

    def __str__(self):
        return '<Process template {!r}>'.format(self.name)


class ProcessTemplateCollection(ObjectTemplateCollection[ProcessTemplate]):
    """A collection of process templates."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/process-templates'
    _dataset_agnostic_path_template = 'projects/{project_id}/process-templates'
    _individual_key = 'process_template'
    _collection_key = 'process_templates'
    _resource = ProcessTemplate

    @classmethod
    def get_type(cls) -> Type[ProcessTemplate]:
        """Return the resource type in the collection."""
        return ProcessTemplate



from citrine.exceptions import NotFound
from citrine.resources import ProjectCollection
from citrine.resources.dataset import Dataset
from citrine.informatics.workflows.design_workflow import DesignWorkflow
from citrine._rest.collection import CreationType, Collection
from copy import deepcopy


def find_collection(*, collection, name):
    """
    Looks through the pages of a collection for a resource with the specified name.

    Returns it, or if not found, returns None
    """
    if isinstance(collection, ProjectCollection):
        try:
            # try to use search if it is available
            # call list() to collapse the iterator, otherwise the NotFound
            # won't show up until collection_list is used
            collection_list = list(collection.search(search_params={
                "name": {
                    "value": name,
                    "search_method": "EXACT"
                }
            }))
        except (NotFound, NotImplementedError):
            # Search must not be available yet or any more
            collection_list = collection.list()
    else:
        collection_list = collection.list()

    matching_resources = [resource for resource in collection_list if resource.name == name]
    if len(matching_resources) > 1:
        raise ValueError("Found multiple collections with name '{}'".format(name))
    if len(matching_resources) == 1:
        result = matching_resources.pop()
        print('Found existing: {}'.format(result))
        return result
    else:
        return None


def get_by_name_or_create(*, collection, name, default_provider):
    """
    Tries to find a collection by its name (returns first hit).

    If not found, implements default_provider
    """
    found = find_collection(collection=collection, name=name)
    if found:
        return found
    else:
        print('Failed to find resource with name {}, creating one instead.'.format(name))
        return default_provider()


def get_by_name_or_raise_error(*, collection, name):
    """
    Tries to find a collection by its name (returns first hit).

    If not found, raises error
    """
    found = find_collection(collection=collection, name=name)
    if found:
        return found
    else:
        raise ValueError("Did not find resource with the given name: {}".format(name))


def find_or_create_project(*, project_collection, project_name, raise_error=False):
    """
    Tries to find a project by name (returns first hit).

    If not found, creates a new project with the given name
    """
    if raise_error:
        project = get_by_name_or_raise_error(collection=project_collection, name=project_name)
    else:
        project = get_by_name_or_create(
            collection=project_collection,
            name=project_name,
            default_provider=lambda: project_collection.register(project_name)
        )
    return project


def find_or_create_team(*, team_collection, team_name, raise_error=False):
    """
    Tries to find a team by name (returns first hit).

    If not found, creates a new team with the given name
    """
    if raise_error:
        team = get_by_name_or_raise_error(collection=team_collection, name=team_name)
    else:
        team = get_by_name_or_create(
            collection=team_collection,
            name=team_name,
            default_provider=lambda: team_collection.register(team_name)
        )
    return team


def find_or_create_dataset(*, dataset_collection, dataset_name, raise_error=False):
    """
    Tries to find a dataset by name (returns first hit).

    If not found, creates a new dataset with the given name
    """
    if raise_error:
        dataset = get_by_name_or_raise_error(collection=dataset_collection, name=dataset_name)
    else:
        dataset = get_by_name_or_create(
            collection=dataset_collection,
            name=dataset_name,
            default_provider=lambda: dataset_collection.register(
                Dataset(dataset_name, summary="seed summ.", description="seed desc.")
            )
        )
    return dataset


def create_or_update(*, collection: Collection[CreationType], resource: CreationType)\
        -> CreationType:
    """
    Update a resource of a given name belonging to a collection.

    Create it if it doesn't exist. If collection already contains
    a resource with the same name it will be updated. If there are multiple
    resources with the same name it will throw an error.

    Parameters
    ----------
    collection : Collection[CreationType]
        Collection within which you want to update or create a resource
    resource : CreationType
        Resource that you want to create or update.

    Returns
    -------
    CreationType
        Registered updated or created resource.

    """
    old_resource = find_collection(collection=collection, name=resource.name)
    if old_resource:
        print("Updating module: {}".format(resource.name))
        # Copy so that passed-in resource is unaffected
        new_resource = deepcopy(resource)
        new_resource.uid = old_resource.uid
        # Locally created design workflows likely won't have a branch ID but
        # need one to be updated.
        if isinstance(old_resource, DesignWorkflow):
            new_resource.branch_id = old_resource.branch_id
        return collection.update(new_resource)
    else:
        print("Registering new module:  {}".format(resource.name))
        return collection.register(resource)



"""Resources that represent collections of modules."""
from uuid import UUID
from typing import TypeVar, Union

from citrine._rest.collection import Collection
from citrine._session import Session
from citrine._utils.functions import migrate_deprecated_argument
from citrine.exceptions import CitrineException
from citrine.informatics.modules import Module


class ModuleCollection(Collection[Module]):
    """Represents the collection of all modules as well as the resources belonging to it.

    Parameters
    ----------
    project_id: UUID
        the UUID of the project

    """

    _path_template = '/projects/{project_id}/modules'
    _individual_key = None
    _resource = Module

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session

    def build(self, data: dict) -> Module:
        """Build an individual module."""
        module = Module.build(data)
        module.session = self.session
        return module


ModuleType = TypeVar('ModuleType', bound='Module')


class AbstractModuleCollection(Collection[ModuleType]):
    """Abstract class for representing collection of modules."""

    _module_type = None

    def delete(self, uid: Union[UUID, str]):
        """Modules cannot be deleted at this time."""
        msg = "Modules cannot be deleted at this time. Use 'archive' instead."
        raise NotImplementedError(msg)

    def archive(self, uid: Union[UUID, str] = None,
                module_id: Union[UUID, str] = None) -> ModuleType:
        """Archiving a module removes it from view, but is not a hard delete.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the module to archive
        module_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        """
        uid = migrate_deprecated_argument(uid, "uid", module_id, "module_id")
        try:
            module = self.get(uid)
        except CitrineException:
            raise RuntimeError(f"{self._module_type} with id {module_id} was not found, "
                               f"and hence cannot be archived.")
        module.archived = True
        return self.update(module)

    def restore(self, uid: Union[UUID, str] = None,
                module_id: Union[UUID, str] = None) -> ModuleType:
        """Restore an archived module.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the module to restore
        module_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        """
        uid = migrate_deprecated_argument(uid, "uid", module_id, "module_id")
        try:
            module = self.get(uid)
        except CitrineException:
            raise RuntimeError(f"{self._module_type} with id {module_id} was not found, "
                               f"and hence cannot be restored.")
        module.archived = False
        return self.update(module)



from typing import Optional


class AsynchronousObject:
    """Abstract class for objects that are asynchronously populated."""

    _in_progress_statuses = []  # list of statuses that indicate 'in progress'
    _succeeded_statuses = []  # list of statuses that indicate 'succeeded'
    _failed_statuses = []  # list of statuses that indicate 'failed'

    def _fetch_status(self) -> Optional[str]:
        """Fetch the status of the resource.

        This is is the `status` attribute by default, but can be overridden if the status
        is encoded differently and/or we want to pull an update from the backend.
        """
        return self.status

    def in_progress(self) -> bool:
        """Whether the backend process is in progress."""
        updated_status = self._fetch_status()
        return any(updated_status == s for s in self._in_progress_statuses)

    def succeeded(self) -> bool:
        """Whether the backend process has completed successfully."""
        updated_status = self._fetch_status()
        return any(updated_status == s for s in self._succeeded_statuses)

    def failed(self) -> bool:
        """Whether the backend process has completed unsuccessfully."""
        updated_status = self._fetch_status()
        return any(updated_status == s for s in self._failed_statuses)



"""Top-level class for all object template objects and collections thereof."""
from abc import ABC
from typing import TypeVar

from citrine.resources.templates import Template, TemplateCollection


class ObjectTemplate(Template, ABC):
    """
    An abstract object template object.

    ObjectTemplate must be extended along with `Resource`
    """


ObjectTemplateResourceType = TypeVar("ObjectTemplateResourceType", bound="ObjectTemplate")


class ObjectTemplateCollection(TemplateCollection[ObjectTemplateResourceType], ABC):
    """A collection of one kind of object template object."""



"""Resources that represent both individual and collections of design workflow executions."""
import sys
from typing import Optional, Union, Iterator
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._utils.functions import shadow_classes_in_module
from citrine._session import Session
import citrine.informatics.executions.design_execution
from citrine.informatics.executions import DesignExecution
from citrine.informatics.scores import Score
from citrine.resources.response import Response


shadow_classes_in_module(citrine.informatics.executions.design_execution, sys.modules[__name__])


class DesignExecutionCollection(Collection["DesignExecution"]):
    """A collection of DesignExecutions."""

    _path_template = '/projects/{project_id}/design-workflows/{workflow_id}/executions'  # noqa
    _individual_key = None
    _collection_key = 'response'
    _resource = DesignExecution

    def __init__(self,
                 project_id: UUID,
                 session: Session,
                 workflow_id: Optional[UUID] = None):
        self.project_id: UUID = project_id
        self.session: Session = session
        self.workflow_id: UUID = workflow_id

    def build(self, data: dict) -> DesignExecution:
        """Build an individual DesignWorkflowExecution."""
        execution = DesignExecution.build(data)
        execution._session = self.session
        execution.project_id = self.project_id
        return execution

    def trigger(self, execution_input: Score):
        """Trigger a Design Workflow execution given a score."""
        path = self._get_path()
        data = self.session.post_resource(path, {'score': execution_input.dump()})
        return self.build(data)

    def register(self, model: DesignExecution) -> DesignExecution:
        """Cannot register an execution."""
        raise NotImplementedError("Cannot register a DesignExecution.")

    def update(self, model: DesignExecution) -> DesignExecution:
        """Cannot update an execution."""
        raise NotImplementedError("Cannot update a DesignExecution.")

    def archive(self, uid: Union[UUID, str]):
        """Archive a Design Workflow execution.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the execution to archive

        """
        raise NotImplementedError(
            "Design Executions cannot be archived")

    def restore(self, uid: UUID):
        """Restore an archived Design Workflow execution.

        Parameters
        ----------
        uid: UUID
            Unique identifier of the execution to restore

        """
        raise NotImplementedError(
            "Design Executions cannot be restored")

    def list(self, *,
             page: Optional[int] = None,
             per_page: int = 100,
             ) -> Iterator[DesignExecution]:
        """
        Paginate over the elements of the collection.

        Leaving page and per_page as default values will yield all elements in the
        collection, paginating over all available pages.

        Parameters
        ---------
        page: int, optional
            The "page" of results to list. Default is to read all pages and yield
            all results.  This option is deprecated.
        per_page: int, optional
            Max number of results to return per page. Default is 100.  This parameter
            is used when making requests to the backend service.  If the page parameter
            is specified it limits the maximum number of elements in the response.
        predictor_id: uuid, optional
            list executions that targeted the predictor with this id

        Returns
        -------
        Iterator[ResourceType]
            Resources in this collection.

        """
        return self._paginator.paginate(page_fetcher=self._fetch_page,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Design Workflow Executions cannot be deleted or archived."""
        raise NotImplementedError(
            "Design Executions cannot be deleted")



import functools
from typing import Iterator, Optional, Union
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.resources.design_workflow import DesignWorkflowCollection


class Branch(Resource['Branch']):
    """
    A project branch.

    A branch is a container for design workflows.
    """

    name = properties.String('name')
    uid = properties.Optional(properties.UUID(), 'id')
    archived = properties.Boolean('archived', serializable=False)
    created_at = properties.Optional(properties.Datetime(), 'created_at', serializable=False)
    updated_at = properties.Optional(properties.Datetime(), 'updated_at', serializable=False)

    project_id: Optional[UUID] = None

    def __init__(self,
                 name: str,
                 *,
                 session: Optional[Session] = None):
        self.name: str = name
        self.session: Session = session

    def __str__(self):
        return f'<Branch {self.name!r}>'

    @property
    def design_workflows(self) -> DesignWorkflowCollection:
        """Return a resource representing all workflows contained within this branch."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot initialize workflow without project reference!')
        return DesignWorkflowCollection(project_id=self.project_id,
                                        session=self.session,
                                        branch_id=self.uid)


class BranchCollection(Collection[Branch]):
    """A collection of Branches."""

    _path_template = '/projects/{project_id}/design-workflow-branches'
    _individual_key = None
    _collection_key = 'response'
    _resource = Branch

    def __init__(self, project_id: UUID, session: Session):
        self.project_id: UUID = project_id
        self.session: Session = session

    def build(self, data: dict) -> Branch:
        """
        Build an individual branch from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the branch.

        Return
        -------
        Branch
            The branch created from data.

        """
        branch = Branch.build(data)
        branch.session = self.session
        branch.project_id = self.project_id
        return branch

    def list(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List active branches using pagination.

        Yields all active branches, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            Active branches in this collection.

        """
        return super().list(per_page=per_page)

    def list_archived(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List archived branches using pagination.

        Yields all archived branches, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            Archived branches in this collection.

        """
        fetcher = functools.partial(self._fetch_page, additional_params={"archived": True})
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def archive(self, uid: Union[UUID, str] = None):
        """
        Archive a branch.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the branch to archive

        """
        archive_path_template = f'{self._get_path(uid)}/archive'
        url = format_escaped_url(archive_path_template, project_id=self.project_id, branch_id=uid)
        data = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(data)

    def restore(self, uid: Union[UUID, str] = None):
        """
        Restore an archived branch.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the branch to restore

        """
        restore_path_template = f'{self._get_path(uid)}/restore'
        url = format_escaped_url(restore_path_template, project_id=self.project_id, branch_id=uid)
        data = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(data)



import platform
from datetime import datetime, timedelta
from json.decoder import JSONDecodeError
from logging import getLogger
from os import environ
from typing import Optional, Callable, Iterator
from urllib.parse import urlunsplit
from warnings import warn

import jwt
import requests
import requests.auth
from requests import Response
from urllib3.util.retry import Retry

import citrine
from citrine._utils.functions import format_escaped_url
from citrine.exceptions import (
    BadRequest,
    CitrineException,
    Conflict,
    NotFound,
    Unauthorized,
    UnauthorizedRefreshToken,
    WorkflowNotReadyException)

# Choose a 5 second buffer so that there's no chance of the access token
# expiring during the check for expiration
EXPIRATION_BUFFER_MILLIS: timedelta = timedelta(milliseconds=5000)
logger = getLogger(__name__)


class Session(requests.Session):
    """Wrapper around requests.Session that is both refresh-token and schema aware."""

    def __init__(self,
                 refresh_token: str = None,
                 legacy_scheme: Optional[str] = None,
                 host: str = None,
                 port: Optional[str] = None,
                 *,
                 scheme: str = None):
        super().__init__()
        if refresh_token is None:
            refresh_token = environ.get('CITRINE_API_KEY')
        if legacy_scheme is not None:
            warn("Creating a session with positional arguments other than refresh_token "
                 "is deprecated; use keyword arguments to specify scheme, host and port.",
                 DeprecationWarning)
            if scheme is None:
                scheme = legacy_scheme
            else:
                raise ValueError("Specify legacy_scheme or scheme, not both.")
        elif scheme is None:
            scheme = 'https'
        if host is None:
            host = environ.get('CITRINE_API_HOST')
            if host is None:
                raise ValueError("No host passed and environmental "
                                 "variable CITRINE_API_HOST not set.")

        self.scheme: str = scheme
        self.authority = ':'.join(([host] if host else []) + ([port] if port else []))
        self.refresh_token: str = refresh_token
        self.access_token: Optional[str] = None
        self.access_token_expiration: datetime = datetime.utcnow()
        self._accounts_service_v3: bool = False

        agent = "{}/{} python-requests/{} citrine-python/{}".format(
            platform.python_implementation(),
            platform.python_version(),
            requests.__version__,
            citrine.__version__)

        # Following scheme:[//authority]path[?query][#fragment] (https://en.wikipedia.org/wiki/URL)
        self.headers.update({
            "Content-Type": "application/json",
            "User-Agent": agent})

        # Default parameters for S3 connectivity. Can be changed by tests.
        self.s3_endpoint_url = None
        self.s3_use_ssl = True
        self.s3_addressing_style = 'auto'

        # Feature flag for enabling the use of Dataset idempotent PUT. Will be removed
        # in a future release.
        self.use_idempotent_dataset_put = False

        # Custom adapter so we can use custom retry parameters. The default HTTP status
        # codes for retries are [503, 413, 429]. We're using status_force list to add
        # additional codes to retry on, focusing on specific CloudFlare 5XX errors.
        retries = Retry(total=10,
                        connect=5,
                        read=5,
                        status=5,
                        backoff_factor=0.25,
                        status_forcelist=[500, 502, 504, 520, 521, 522, 524, 527])
        adapter = requests.adapters.HTTPAdapter(max_retries=retries)
        self.mount('https://', adapter)
        self.mount('http://', adapter)

        # Requests has it's own set of exceptions that do not inherit from the
        # built-in exceptions. The built-in ConnectionError handles 4 different
        # child exceptions: https://docs.python.org/3/library/exceptions.html#ConnectionError
        self.retry_errs = (ConnectionError,
                           requests.exceptions.ConnectionError,
                           requests.exceptions.ChunkedEncodingError)
        self._refresh_access_token()
        self._check_accounts_version()

    def _versioned_base_url(self, version: str = 'v1'):
        return urlunsplit((
            self.scheme,
            self.authority,
            format_escaped_url('api/{}/', version),
            '',  # query string
            ''  # fragment
        ))

    def _is_access_token_expired(self):
        return self.access_token_expiration - EXPIRATION_BUFFER_MILLIS <= datetime.utcnow()

    def _refresh_access_token(self) -> None:
        """Optionally refresh our access token (if the previous one is about to expire)."""
        data = {'refresh_token': self.refresh_token}

        response = self._request_with_retry('POST', self._versioned_base_url() + 'tokens/refresh',
                                            json=data)

        if response.status_code != 200:
            raise UnauthorizedRefreshToken()
        self.access_token = response.json()['access_token']
        self.access_token_expiration = datetime.utcfromtimestamp(
            jwt.decode(self.access_token, verify=False)['exp']
        )

        # Explicitly set an updated 'auth', so as to not rely on implicit cookie handling.
        self.auth = BearerAuth(self.access_token)

    def _check_accounts_version(self) -> None:
        """Checks Product to find out what version of Accounts is used."""
        response = self._request_with_retry('GET',
                                            self._versioned_base_url() + 'utils/runtime-config')

        if response.status_code != 200:
            raise CitrineException(response.text)
        self._accounts_service_v3 = response.json().get('accounts_service_v3', False)

    def _request_with_retry(self, method, uri, **kwargs):
        """Wrap a request with a try/except to retry when ConnectionErrors are seen."""
        # The urllib3 Retry object does not handle retries when ConnectionErrors
        # (or other similar errors) and raised.  Using a stale connection causes
        # these issues.  Retrying the request uses a new connection.  See PLA-3449/4183.
        try:
            response = self.request(method, uri, **kwargs)
        except self.retry_errs as e:
            logger.warning('{} seen, retrying request'.format(repr(e)))
            response = self.request(method, uri, **kwargs)

        return response

    def checked_request(self, method: str, path: str,
                        version: str = 'v1', **kwargs) -> requests.Response:
        """Check response status code and throw an exception if relevant."""
        logger.debug('BEGIN request details:')
        logger.debug('\tmethod: {}'.format(method))
        logger.debug('\tpath: {}'.format(path))
        logger.debug('\tversion: {}'.format(version))

        if self._is_access_token_expired():
            self._refresh_access_token()
            self._check_accounts_version()
        uri = self._versioned_base_url(version) + path.lstrip('/')

        logger.debug('\turi: {}'.format(uri))

        for k, v in kwargs.items():
            logger.debug('\t{}: {}'.format(k, v))
        logger.debug('END request details.')

        response = self._request_with_retry(method, uri, **kwargs)

        try:
            if response.status_code == 401 and response.json().get("reason") == "invalid-token":
                self._refresh_access_token()
                self._check_accounts_version()
                response = self._request_with_retry(method, uri, **kwargs)
        except AttributeError:
            # Catch AttributeErrors and log response
            # The 401 status will be handled further down
            logger.error("Failed to decode json from response: {}".format(response.text))
        except ValueError:
            # Ignore ValueErrors thrown by attempting to decode json bodies. This
            # might occur if we get a 401 response without a JSON body
            pass

        if 200 <= response.status_code <= 299:
            logger.info('%s %s %s', response.status_code, method, path)
            return response
        else:
            stacktrace = self._extract_response_stacktrace(response)
            if stacktrace is not None:
                logger.error('Response arrived with stacktrace:')
                logger.error(stacktrace)
            if response.status_code == 400:
                logger.error('%s %s %s', response.status_code, method, path)
                logger.error(response.text)
                raise BadRequest(path, response)
            elif response.status_code == 401:
                logger.error('%s %s %s', response.status_code, method, path)
                raise Unauthorized(path, response)
            elif response.status_code == 403:
                logger.error('%s %s %s', response.status_code, method, path)
                raise Unauthorized(path, response)
            elif response.status_code == 404:
                logger.error('%s %s %s', response.status_code, method, path)
                raise NotFound(path, response)
            elif response.status_code == 409:
                logger.debug('%s %s %s', response.status_code, method, path)
                raise Conflict(path, response)
            elif response.status_code == 425:
                logger.debug('%s %s %s', response.status_code, method, path)
                msg = 'Cant execute at this time. Try again later. Error: {}'.format(response.text)
                raise WorkflowNotReadyException(msg)
            else:
                logger.error('%s %s %s', response.status_code, method, path)
                raise CitrineException(response.text)

    @staticmethod
    def _extract_response_stacktrace(response: Response) -> Optional[str]:
        try:
            json_value = response.json()
            if isinstance(json_value, dict):
                return json_value.get('debug_stacktrace')
        except ValueError:
            pass
        return None

    def get_resource(self, path: str, **kwargs) -> dict:
        """GET a particular resource as JSON."""
        response = self.checked_get(path, **kwargs)
        return self._extract_response_json(path, response)

    def post_resource(self, path: str, json: dict, **kwargs) -> dict:
        """POST to a particular resource as JSON."""
        response = self.checked_post(path, json=json, **kwargs)
        return self._extract_response_json(path, response)

    def put_resource(self, path: str, json: dict, **kwargs) -> dict:
        """PUT data given by some JSON at a particular resource."""
        response = self.checked_put(path, json=json, **kwargs)
        return self._extract_response_json(path, response)

    def patch_resource(self, path: str, json: dict, **kwargs) -> dict:
        """PATCH data given by some JSON at a particular resource."""
        response = self.checked_patch(path, json=json, **kwargs)
        return self._extract_response_json(path, response)

    def delete_resource(self, path: str, **kwargs) -> dict:
        """DELETE a particular resource as JSON."""
        response = self.checked_delete(path, **kwargs)
        return self._extract_response_json(path, response)

    @staticmethod
    def _extract_response_json(path, response) -> dict:
        """Extract json from the response or log and return an empty dict if extraction fails."""
        extracted_response = {}
        try:
            if "application/json" in response.headers.get("Content-Type", ""):
                extracted_response = response.json()
            else:  # pragma: no cover
                logger.info(f"""Response at {path} with status code of {response.status_code}
                    lacked the required 'application/json' Content-Type in the header.""")

        except JSONDecodeError as err:
            logger.info('Response at path %s with status code %s failed json parsing with'
                        ' exception %s. Returning empty value.',
                        path,
                        response.status_code,
                        err.msg)

        return extracted_response

    @staticmethod
    def cursor_paged_resource(base_method: Callable[..., dict], path: str,
                              forward: bool = True, per_page: int = 100,
                              version: str = 'v2', **kwargs) -> Iterator[dict]:
        """
        Returns a flat generator of results for an API query.

        Results are fetched in chunks of size `per_page` and loaded lazily.
        """
        params = kwargs.get('params', {})
        params['forward'] = forward
        params['ascending'] = forward
        params['per_page'] = per_page
        kwargs['params'] = params
        while True:
            response_json = base_method(path, version=version, **kwargs)
            for obj in response_json['contents']:
                yield obj
            cursor = response_json.get('next')
            if cursor is None:
                break
            params['cursor'] = cursor

    def checked_post(self, path: str, json: dict, **kwargs) -> Response:
        """Execute a POST request to a URL and utilize error filtering on the response."""
        return self.checked_request('POST', path, json=json, **kwargs)

    def checked_put(self, path: str, json: dict, **kwargs) -> Response:
        """Execute a PUT request to a URL and utilize error filtering on the response."""
        return self.checked_request('PUT', path, json=json, **kwargs)

    def checked_patch(self, path: str, json: dict, **kwargs) -> Response:
        """Execute a PATCH request to a URL and utilize error filtering on the response."""
        return self.checked_request('PATCH', path, json=json, **kwargs)

    def checked_delete(self, path: str, **kwargs) -> Response:
        """Execute a DELETE request to a URL and utilize error filtering on the response."""
        return self.checked_request('DELETE', path, **kwargs)

    def checked_get(self, path: str, **kwargs) -> Response:
        """Execute a GET request to a URL and utilize error filtering on the response."""
        return self.checked_request('GET', path, **kwargs)


class BearerAuth(requests.auth.AuthBase):
    """A lightweight Auth class to support Bearer tokens."""

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        """Generate the appropriate Authorization header."""
        r.headers["Authorization"] = "Bearer " + self.token
        return r



"""Resources that represent collections of predictors."""
from functools import partial
from uuid import UUID
from typing import Iterable, TypeVar, Optional, Union

from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._session import Session
from citrine._utils.functions import migrate_deprecated_argument, format_escaped_url
from citrine.exceptions import Conflict
from citrine.resources.module import AbstractModuleCollection
from citrine.informatics.data_sources import DataSource
from citrine.informatics.predictors import Predictor

CreationType = TypeVar('CreationType', bound=Predictor)


class AutoConfigureMode(BaseEnumeration):
    """[ALPHA] The format to use in building auto-configured assets.

    * PLAIN corresponds to a single-row GEM table and plain predictor
    * FORMULATION corresponds to a multi-row GEM table and formulations predictor
    * INFER auto-detects the GEM table and predictor type
    """

    PLAIN = 'PLAIN'
    FORMULATION = 'FORMULATION'
    INFER = 'INFER'


class PredictorCollection(AbstractModuleCollection[Predictor]):
    """Represents the collection of all predictors for a project.

    Parameters
    ----------
    project_id: UUID
        the UUID of the project

    """

    _api_version = 'v3'
    _path_template = '/projects/{project_id}/predictors'
    _individual_key = None
    _resource = Predictor
    _collection_key = 'response'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session

    def _predictors_path(self, subpath: str, uid: Union[UUID, str] = None):
        path_template = self._path_template + (f"/{uid}" if uid else "") + f"/{subpath}"
        return format_escaped_url(path_template, project_id=self.project_id)

    def build(self, data: dict) -> Predictor:
        """Build an individual Predictor."""
        predictor: Predictor = Predictor.build(data)
        predictor._session = self.session
        predictor._project_id = self.project_id
        return predictor

    def register(self, predictor: Predictor) -> Predictor:
        """Register and train a Predictor."""
        created_predictor = super().register(predictor)

        # We never exposed saving a model without training, so we should1
        # continue to do it automatically.
        return self._train(created_predictor.uid)

    def update(self, predictor: Predictor) -> Predictor:
        """Update and train a Predictor."""
        super().update(predictor)

        # The /api/v3/predictors endpoint switched (un)archive from a field on the update payload
        # to their own endpoints. To maintain backwards compatibilty, all predictors have an
        # _archived field set by the archived property. It will be archived if True, and restored
        # if False. It defaults to None, which does nothing. The value is reset afterwards.
        if predictor._archived is True:
            self.archive(predictor.uid)
        elif predictor._archived is False:
            self.restore(predictor.uid)
        predictor._archived = None

        # We never exposed saving a model without training, so we should
        # continue to do it automatically.
        return self._train(predictor.uid)

    def _train(self, uid: Union[UUID, str]):
        path = self._predictors_path("train", uid)
        entity = self.session.put_resource(path, {}, version=self._api_version)
        return self.build(entity)

    def archive(self, uid: Union[UUID, str]):
        """Archive a predictor.

        uid: Union[UUID, str]
            Unique identifier of the predictor to archive.

        """
        path = self._predictors_path("archive", uid)
        entity = self.session.put_resource(path, {}, version=self._api_version)
        return self.build(entity)

    def restore(self, uid: Union[UUID, str]):
        """Restore an archived predictor.

        uid: Union[UUID, str]
            Unique identifier of the predictor to restore.

        """
        path = self._predictors_path("restore", uid)
        entity = self.session.put_resource(path, {}, version=self._api_version)
        return self.build(entity)

    def list_archived(self,
                      *,
                      page: Optional[int] = None,
                      per_page: int = 20) -> Iterable[Predictor]:
        """List archived Predictors."""
        fetcher = partial(self._fetch_page, additional_params={"filter": "archived eq 'true'"})
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def check_for_update(self, uid: Union[UUID, str] = None,
                         predictor_id: Union[UUID, str] = None) -> Optional[Predictor]:
        """
        Check if there are updates available for a predictor.

        Typically these are updates to the training data. For example, a GEM Table may have
        been re-built to include additional rows.

        This check does not update the predictor; it just returns the update that is available.
        To perform the update, the response should then be used to call PredictorCollection.update

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the predictor to check
        predictor_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        Returns
        -------
        Optional[Predictor]
            The update, if an update is available; None otherwise.

        """
        uid = migrate_deprecated_argument(uid, "uid", predictor_id, "predictor_id")
        path = self._predictors_path("update-check", uid)
        update_data = self.session.get_resource(path, version=self._api_version)
        if update_data["updatable"]:
            built = Predictor.build(update_data)
            built.uid = uid
            return built
        else:
            return None

    def auto_configure(self, *, training_data: DataSource,
                       pattern=AutoConfigureMode.INFER, prefer_valid=True) -> Predictor:
        """[ALPHA] Create a default predictor for some training data.

        This method will return an unregistered predictor generated by inspecting the
        training data and attempting to automatically configure the predictor.

        The configuration generated while using the `AutoConfigureMode.SIMPLE` pattern
        includes featurizers for chemical formulas/molecular structures,
        and `AutoMLPredictor`s for any variables identified as responses in the training data.
        The configuration generated while using the `AutoConfigureMode.FORMULATION` pattern
        includes these same components, as well as a `SimpleMixturePredictor`,
        `LabelFractionsPredictor`, `IngredientFractionsPredictor`, and a series of
        `MeanPropertyPredictor`s to handle featurization of formulation quantities
        and ingredient properties.
        The `AutoConfigureMode.INFER` pattern chooses an appropriate mode based on whether
        the data source contains formulations data or not.

        Parameters
        ----------
        training_data: DataSource
            The data to configure the predictor to model.
        pattern: AutoConfigureMode or str
            The predictor pattern to use, either "PLAIN", "FORMULATION", or "INFER".
            The "INFER" pattern auto-detects whether the `DataSource` contains formulations
            data or not.
            If it does, then a formulation predictor is created.
            If not, then a plain predictor is created.
        prefer_valid: Boolean
            If True, enables filtering of sparse descriptors and trimming of
            excess graph components in attempt to return a default configuration
            that will pass validation.
            Default: True.

        Returns
        -------
        Predictor
            Automatically configured predictor for the training data

        """
        # Continue handling string pattern inputs
        if not isinstance(pattern, AutoConfigureMode):
            pattern = AutoConfigureMode.get_enum(pattern)
        pattern = pattern.value

        path = self._predictors_path("default")
        body = {"data_source": training_data.dump(), "pattern": pattern,
                "prefer_valid": prefer_valid}
        data = self.session.post_resource(path, json=body, version=self._api_version)
        return self.build(Predictor.wrap_instance(data["instance"]))

    def convert_to_graph(self, uid: Union[UUID, str], retrain_if_needed: bool = False):
        """Given a SimpleML or Graph predictor, get an equivalent Graph predictor.

        Returns a Graph predictor with any SimpleML predictors converted to an equivalent AutoML
        predictor. If it's not a SimpleML or Graph predictor, or it's not in the READY state, an
        error is raised. SimpleML predictors are deprecated, so this is to aid in your migration.

        Note this conversion is not performed in place! That is, the predictor returned is not
        persisted on the platform. To persist it, pass the converted predictor to
        :py:meth:`citrine.resources.project.PredictorCollection.update`. Or you can do this in
        one step with :py:meth:`citrine.resources.project.PredictorCollection.convert_and_update`.

        .. code:: python

            converted = project.predictors.convert_to_graph(predictor_uid)
            project.predictors.update(converted)

            # is equivalent to

            converted = project.predictors.convert_and_update(predictor_uid)

        If a predictor needs to be retrained before conversion, it will raise an HTTP 409 Conflict
        error. This may occur when the Citrine Platform has been updated since your predictor was
        last trained. If `retrain_if_needed` is `True`, retraining will automatically begin, and
        the method will return `None. Once retraining completes, call this method again to get the
        converted predictor. For example:

        .. code:: python

            converted = project.predictors.convert_and_update(pred.uid, retrain_if_needed=True)
            if converted is None:
                predictor = project.predictors.get(pred.uid)
                wait_while_validating(collection=project.predictors, module=predictor)
                converted = project.predictors.convert_and_update(pred.uid)
        """
        path = self._predictors_path("convert", uid)
        try:
            entity = self.session.get_resource(path, version=self._api_version)
            return self.build(entity)
        except Conflict as exc:
            if retrain_if_needed:
                self.update(self.get(uid))
                return None
            else:
                raise exc

    def convert_and_update(self, uid: Union[UUID, str], retrain_if_needed: bool = False):
        """Given a SimpleML or Graph predictor, overwrite it with an equivalent Graph predictor.

        See `PredictorCollection.convert_to_graph` for more detail.
        """
        new_pred = self.convert_to_graph(uid, retrain_if_needed)
        return self.update(new_pred) if new_pred else None
